/*
 * This file is a component of 3wks Analytics, a software library from 3wks.
 * Copyright (C) 2013 3wks, <support@3wks.com.au>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.threewks.analytics.filter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;

import org.joda.time.DateTime;

/**
 * A servlet response wrapper that captures headers, cookies, status code and redirect URL for reporting purposes.
 * This class is used by {@link HttpRequestResponseTrackingFilter} to capture information about the response.
 */
public class AnalyticsHttpServletResponseWrapper extends HttpServletResponseWrapper {

	private Map<String, List<String>> headers = new HashMap<String, List<String>>();
	private List<Cookie> cookies = new ArrayList<Cookie>();
	private int statusCode = SC_OK;
	private String redirectUrl;

	public AnalyticsHttpServletResponseWrapper(HttpServletResponse response) {
		super(response);
	}

	@Override
	public void addCookie(Cookie cookie) {
		super.addCookie(cookie);
		cookies.add(cookie);
	}

	@Override
	public void sendError(int sc, String msg) throws IOException {
		super.sendError(sc, msg);
		this.statusCode = sc;
	}

	@Override
	public void sendError(int sc) throws IOException {
		super.sendError(sc);
		this.statusCode = sc;
	}

	@Override
	public void setStatus(int sc) {
		super.setStatus(sc);
		this.statusCode = sc;
	}

	@Override
	public void setStatus(int sc, String sm) {
		super.setStatus(sc, sm);
		this.statusCode = sc;
	}

	@Override
	public void setHeader(String name, String value) {
		super.setHeader(name, value);
		setSavedHeader(name, value);
	}

	@Override
	public void addHeader(String name, String value) {
		super.addHeader(name, value);
		addSavedHeader(name, value);
	}

	@Override
	public void setDateHeader(String name, long date) {
		super.setDateHeader(name, date);
		setSavedHeader(name, new DateTime(date).toString());
	}

	@Override
	public void addDateHeader(String name, long date) {
		super.addDateHeader(name, date);
		addSavedHeader(name, new DateTime(date).toString());
	}

	@Override
	public void setIntHeader(String name, int value) {
		super.setIntHeader(name, value);
		setSavedHeader(name, String.valueOf(value));
	}

	@Override
	public void addIntHeader(String name, int value) {
		super.addIntHeader(name, value);
		addSavedHeader(name, String.valueOf(value));
	}

	@Override
	public void sendRedirect(String location) throws IOException {
		super.sendRedirect(location);
		this.redirectUrl = location;
	}

	/**
	 * Save a header value by overwriting any previously saved values.
	 * 
	 * @param name the header name.
	 * @param value the header value.
	 */
	private void setSavedHeader(String name, String value) {
		headers.put(name, Arrays.asList(value));
	}

	/**
	 * Save a header value by adding to the existing values.
	 * 
	 * @param name the header name.
	 * @param value the header value.
	 */
	private void addSavedHeader(String name, String value) {
		List<String> values = headers.get(name);
		if (values == null) {
			values = new ArrayList<String>();
			headers.put(name, values);
		}
		values.add(value);
	}

	public List<Cookie> getCookies() {
		return cookies;
	}

	public Map<String, List<String>> getHeaders() {
		return headers;
	}

	public int getStatusCode() {
		return statusCode;
	}

	public String getRedirectUrl() {
		return redirectUrl;
	}

}
