/*
 * This file is a component of 3wks Analytics, a software library from 3wks.
 * Copyright (C) 2013 3wks, <support@3wks.com.au>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.threewks.analytics.model;

import java.security.Principal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;

import com.threewks.analytics.Util;

public class HttpRequest {

	private String id;
	private String timestamp;
	private String authType;
	private String characterEncoding;
	private int contentLength;
	private String contentType;
	private String contextPath;
	private String localAddr;
	private String localName;
	private int localPort;
	private String country;
	private String language;
	private String method;
	private String pathInfo;
	private String protocol;
	private String queryString;
	private String remoteAddr;
	private String remoteHost;
	private int remotePort;
	private String remoteUser;
	private String requestURI;
	private String requestURL;
	private String scheme;
	private String serverName;
	private int serverPort;
	private String sessionId;
	private String user;

	List<HttpRequestParameter> parameters = new ArrayList<HttpRequestParameter>();
	List<HttpHeader> headers = new ArrayList<HttpHeader>();
	List<HttpCookie> cookies = new ArrayList<HttpCookie>();

	/**
	 * Create a new HttpRequest with a randomly generated id and a timestamp set to the current time.
	 */
	public HttpRequest() {
		this.id = Util.randomId();
		this.timestamp = Util.timestamp();
	}

	/**
	 * Create a new HttpRequest from the given {@link HttpServletRequest} and with a randomly generated id
	 * and a timestamp set to the current time.
	 * 
	 * @param httpServletRequest the HTTP servlet request to extract the data from.
	 */
	public HttpRequest(HttpServletRequest httpServletRequest) {
		this();

		authType = httpServletRequest.getAuthType();
		characterEncoding = httpServletRequest.getCharacterEncoding();
		contentLength = httpServletRequest.getContentLength();
		contentType = httpServletRequest.getContentType();
		contextPath = httpServletRequest.getContextPath();
		localAddr = httpServletRequest.getLocalAddr();
		localName = httpServletRequest.getLocalName();
		localPort = httpServletRequest.getLocalPort();
		method = httpServletRequest.getMethod();
		pathInfo = httpServletRequest.getPathInfo();
		protocol = httpServletRequest.getProtocol();
		queryString = httpServletRequest.getQueryString();
		remoteAddr = httpServletRequest.getRemoteAddr();
		remoteHost = httpServletRequest.getRemoteHost();
		remotePort = httpServletRequest.getRemotePort();
		remoteUser = httpServletRequest.getRemoteUser();
		requestURI = httpServletRequest.getRequestURI();
		requestURL = httpServletRequest.getRequestURL().toString();
		scheme = httpServletRequest.getScheme();
		serverName = httpServletRequest.getServerName();
		serverPort = httpServletRequest.getServerPort();
		sessionId = httpServletRequest.getRequestedSessionId();

		Principal userPrincipal = httpServletRequest.getUserPrincipal();
		user = userPrincipal != null ? userPrincipal.getName() : null;

		Locale locale = httpServletRequest.getLocale();
		country = locale.getDisplayCountry();
		language = locale.getDisplayLanguage();

		extractParameters(httpServletRequest, parameters);
		extractHeaders(httpServletRequest, headers);
		extractCookies(httpServletRequest, cookies);
	}

	public String getId() {
		return id;
	}

	public void setId(String id) {
		this.id = id;
	}

	public String getTimestamp() {
		return timestamp;
	}

	public void setTimestamp(String timestamp) {
		this.timestamp = timestamp;
	}

	public String getAuthType() {
		return authType;
	}

	public void setAuthType(String authType) {
		this.authType = authType;
	}

	public String getProtocol() {
		return protocol;
	}

	public void setProtocol(String protocol) {
		this.protocol = protocol;
	}

	public String getCharacterEncoding() {
		return characterEncoding;
	}

	public void setCharacterEncoding(String characterEncoding) {
		this.characterEncoding = characterEncoding;
	}

	public int getContentLength() {
		return contentLength;
	}

	public void setContentLength(int contentLength) {
		this.contentLength = contentLength;
	}

	public String getContentType() {
		return contentType;
	}

	public void setContentType(String contentType) {
		this.contentType = contentType;
	}

	public String getContextPath() {
		return contextPath;
	}

	public void setContextPath(String contextPath) {
		this.contextPath = contextPath;
	}

	public String getLocalAddr() {
		return localAddr;
	}

	public void setLocalAddr(String localAddr) {
		this.localAddr = localAddr;
	}

	public String getLocalName() {
		return localName;
	}

	public void setLocalName(String localName) {
		this.localName = localName;
	}

	public int getLocalPort() {
		return localPort;
	}

	public void setLocalPort(int localPort) {
		this.localPort = localPort;
	}

	public String getCountry() {
		return country;
	}

	public void setCountry(String country) {
		this.country = country;
	}

	public String getLanguage() {
		return language;
	}

	public void setLanguage(String language) {
		this.language = language;
	}

	public String getMethod() {
		return method;
	}

	public void setMethod(String method) {
		this.method = method;
	}

	public String getPathInfo() {
		return pathInfo;
	}

	public void setPathInfo(String pathInfo) {
		this.pathInfo = pathInfo;
	}

	public String getQueryString() {
		return queryString;
	}

	public void setQueryString(String queryString) {
		this.queryString = queryString;
	}

	public String getRemoteAddr() {
		return remoteAddr;
	}

	public void setRemoteAddr(String remoteAddr) {
		this.remoteAddr = remoteAddr;
	}

	public String getRemoteHost() {
		return remoteHost;
	}

	public void setRemoteHost(String remoteHost) {
		this.remoteHost = remoteHost;
	}

	public int getRemotePort() {
		return remotePort;
	}

	public void setRemotePort(int remotePort) {
		this.remotePort = remotePort;
	}

	public String getRemoteUser() {
		return remoteUser;
	}

	public void setRemoteUser(String remoteUser) {
		this.remoteUser = remoteUser;
	}

	public String getRequestURI() {
		return requestURI;
	}

	public void setRequestURI(String requestURI) {
		this.requestURI = requestURI;
	}

	public String getRequestURL() {
		return requestURL;
	}

	public void setRequestURL(String requestURL) {
		this.requestURL = requestURL;
	}

	public String getScheme() {
		return scheme;
	}

	public void setScheme(String scheme) {
		this.scheme = scheme;
	}

	public String getServerName() {
		return serverName;
	}

	public void setServerName(String serverName) {
		this.serverName = serverName;
	}

	public int getServerPort() {
		return serverPort;
	}

	public void setServerPort(int serverPort) {
		this.serverPort = serverPort;
	}

	public String getUser() {
		return user;
	}

	public void setUser(String user) {
		this.user = user;
	}

	public String getSessionId() {
		return sessionId;
	}

	public void setSessionId(String sessionId) {
		this.sessionId = sessionId;
	}

	public List<HttpRequestParameter> getParameters() {
		return parameters;
	}

	public void setParameters(List<HttpRequestParameter> parameters) {
		this.parameters = parameters;
	}

	public List<HttpHeader> getHeaders() {
		return headers;
	}

	public void setHeaders(List<HttpHeader> headers) {
		this.headers = headers;
	}

	public List<HttpCookie> getCookies() {
		return cookies;
	}

	public void setCookies(List<HttpCookie> cookies) {
		this.cookies = cookies;
	}

	/**
	 * Extract the cookies from the tiven {@link HttpServletRequest} and add them to the list of {@link HttpCookie} objects.
	 * 
	 * @param httpServletRequest the request to extract the cookies from.
	 * @param cookies the list of {@link HttpCookie} objects to add the cookies to.
	 */
	private static void extractCookies(HttpServletRequest httpServletRequest, List<HttpCookie> cookies) {
		if (httpServletRequest.getCookies() != null) {
			for (Cookie cookie : httpServletRequest.getCookies()) {
				cookies.add(new HttpCookie(cookie));
			}
		}
	}

	/**
	 * Extract the headers from the given {@link HttpServletRequest} and add them to the list of {@link HttpHeader} objects.
	 * Note that if a header has multiple values they are combined into a single, comma separated list.
	 * 
	 * @param httpServletRequest the request to extract the headers from.
	 * @param headers the list of {@link HttpHeader} objects to add the headers to.
	 */
	@SuppressWarnings("unchecked")
	private static void extractHeaders(HttpServletRequest httpServletRequest, List<HttpHeader> headers) {
		Enumeration<String> headerNames = httpServletRequest.getHeaderNames();
		if (headerNames != null) {
			while (headerNames.hasMoreElements()) {
				String name = headerNames.nextElement();
				List<String> values = Collections.list(httpServletRequest.getHeaders(name));
				String value = values.size() > 1 ? values.toString() : httpServletRequest.getHeader(name);
				headers.add(new HttpHeader(name, value));
			}
		}
	}

	/**
	 * Extract the request parameters from the given {@link HttpServletRequest} and add them to the list of {@link HttpRequestParameter} objects.
	 * Note that if a parameter has multiple values they are combined into a single, comma separated list.
	 * 
	 * @param httpServletRequest the request to extract the headers from.
	 * @param parameters the list of {@link HttpRequestParameter} objects to add the parameters to.
	 */
	@SuppressWarnings("unchecked")
	private static void extractParameters(HttpServletRequest httpServletRequest, List<HttpRequestParameter> parameters) {
		Map<String, String[]> parameterMap = httpServletRequest.getParameterMap();
		for (Map.Entry<String, String[]> entry : parameterMap.entrySet()) {
			String name = entry.getKey();
			List<String> values = Arrays.asList(entry.getValue());
			String value = values.size() > 1 ? values.toString() : httpServletRequest.getParameter(name);
			parameters.add(new HttpRequestParameter(name, value));
		}
	}

}
